#!/usr/bin/python

import signal
import pycurl
import xmltodict
import logging
import platform
from ardi.api import core, subscription, service
from StringIO import StringIO
from subprocess import call
import subprocess
import os
import traceback
import threading
import time
import argparse

def CatchSignal(a,b):
    print "Attempting to Terminate"
    signal.signal(signal.SIGINT,signal.SIG_DFL)
    signal.signal(signal.SIGTERM,signal.SIG_DFL)
    note.Stop()

def processqueue():
    while True:
        note.ProcessQueue();
        time.sleep(1)
    

def remove_html_markup(s):
    tag = False
    quote = False
    out = ""

    for c in s:
            if c == '<' and not quote:
                tag = True
            elif c == '>' and not quote:
                tag = False
            elif (c == '"' or c == "'") and tag:
                quote = not quote
            elif not tag:
                out = out + c

    return out

def CallIt(code,params):
    print "    " + code
    cmdline = "send" + code
    env = {}
    for key in params:
        env[key.upper().encode('ascii','ignore')] = str(params[key]).encode('ascii','ignore')

    rt = os.environ.get('SystemRoot','')
    if rt <> "":
        env['SystemRoot'] = rt

    #logging.debug("Sending Via " + cmdline + "...")
    
    #print str(env)
    print "Calling " + cmdline
    try:        
        if platform.system() == 'Windows':                
            cmdline = os.path.dirname(os.path.realpath(__file__)) + "\\" + cmdline
            DETACHED_PROCESS = 0x00000008
            #print "Calling " + str(['c:\\python27\\python.exe',cmdline + ".py"])
            subprocess.Popen(['c:\\python27\\python.exe',cmdline + ".py"],env=env, shell=False, stdin=None, stdout=None,stderr=None,close_fds=True,creationflags=DETACHED_PROCESS)
            #call(['c:\\python27\\python.exe',cmdline + ".py"],env=env)
        else:
            cmdline = os.path.dirname(os.path.realpath(__file__)) + "/" + cmdline
            cmdline += ".py"
            print "Calling " + cmdline
            subprocess.call([cmdline],env=env)
    except:
        traceback.print_exc()
        print "Failed To Run Notification Process"

def SendNotice(server,noti,value):
    response = StringIO()

    fullurl = server + "/api/notifier/getnoticeinfo?id=" + str(noti) + "&value=" + str(value)
    #print fullurl

    c = pycurl.Curl()
    c.setopt(c.URL,fullurl)
    c.setopt(c.WRITEFUNCTION,response.write)
    c.setopt(c.TIMEOUT,5)
    c.setopt(c.CONNECTTIMEOUT,5)
    c.perform()
    ret = response.getvalue()
    c.close()

    #print ret

    result = xmltodict.parse(ret)

    #print str(result)

    try:
        methset = result['noticeinfo']['methods']['method']
    except:
        #print "Failure: " + str(methset)
        pass

    methodparams = {}
    #print str(methset)

    if '@code' in methset:
        iset = []
        iset.append(result['noticeinfo']['methods']['method'])
    else:
        iset = result['noticeinfo']['methods']['method']

    #print str(iset)

    for d in iset:
        dta = {}
            
        code = d['@code']
        
        for key in d:
            if (key == '@code'):
                continue
            dta[key[1:]] = str(d[key])

        methodparams[str(code)] = dta

    try:
        methset = result['noticeinfo']['contact']
    except:
        return

    try:
        if '@id' in methset:
            iset = []
            iset.append(result['noticeinfo']['contact'])
        else:
            iset = result['noticeinfo']['contact']
    except:
        print "No Contacts For This Notification"
        return
        
    for d in iset:
        id = d['@id']
        name = d['@name']

        print "Contact " + str(name) + " is contactable via..."

        try:
            if '@code' in d['method']:
                dset = []
                dset.append(d['method'])
            else:
                dset = d['method']
        except:
            #print "Ignoring " + name + " ( " + id + " ) - No Contact Methods"
            continue

        for r in dset:
            dta = {}
            
            code = r['@code']
            #print "    " + code
            for key in r:
                if (key == '@code'):
                    continue
                #print "      " + str(key[1:]) + "=" + str(r[key])
                dta[key[1:]] = str(r[key])

            final = {}
            final.update(dta)
            final.update(methodparams[code])
            final['title'] = result['noticeinfo']['notice']["name"]
            final['detail'] = result['noticeinfo']['notice']["desc"]
            final['footer'] = result['noticeinfo']['notice']["footer"]

            CallIt(code,final)

def CallMe(content,context):
    #print str(content)
    context.ProcessNewValues(content)

class Notification:
    def __init__(self):
        self.style = 'change'
        self.id = 0
        self.timer = 0
        self.value = 0
        self.delay = 0

class notice:
    def __init__(self):
        self.address=''
        self.trigger='#'
        self.delay = 0

class channel:
    def __init__(self):
        self.ardi = ""
        self.field = ""

class notifier(service.ARDIService):
    def __init__(self):
        super(notifier,self).__init__()
        self.notices = []
        self.methods = {}
        self.queue = []
        self.queuelock = threading.Lock()
        self.addon = "notifier"
        self.description = "Sends out notifications in response to ARDI changes and alerts."

    def GetConfiguration(self,servername):
        super(notifier,self).GetConfiguration(servername)
        try:
             response = StringIO()
            
             fullurl = "http://" + servername + "/api/notifier/getnotifications?trusted=true"
             self.logger.info("Getting configuration from " + fullurl)
             #fullurl = "http://" + servername + "/"
             #print fullurl
             c = pycurl.Curl()
             c.setopt(c.URL,fullurl)
             c.setopt(c.WRITEFUNCTION,response.write)
             c.setopt(c.TIMEOUT,5)
             c.setopt(c.CONNECTTIMEOUT,5)
             c.perform()
             ret = response.getvalue()
             c.close()
             #print "XML Content: " + ret
             return ret
        except:
             time.sleep(10)
             pass

        return False
    
    def ProcessConfiguration(self,xml):
        """Imports the configuration XML file

        This function loads the list of data points within the XML text provided. The function is usually called from LoadConfig

        \param xml: The configuration data in XML format"""
        super(notifier,self).ProcessConfiguration(xml)
        result = xmltodict.parse(xml)
                        
        #Core Parameters
        #print str(xml)
        
        try:
            if '@name' in result['notices']['notice']:
                pass
        except:
            logging.error('The server currently has no notifications')
            return

        #signal.signal(signal.SIGINT,CatchSignal)
        #signal.signal(signal.SIGTERM,CatchSignal)

        nset = result['notices']['notice']

        if '@name' in result['notices']['notice']:
            #print 'Single Item: ' + str(result['notices']['notice'])
            nset = [nset]
        
        for itm in nset:
            no = notice()
            no.previous = None
            no.id = itm['@id']

            asset = itm['@asset']
            prop = itm['@property']
            trigger = str(itm['@trigger'])

            #This is a 'change in value' trigger...
            if trigger[0] == '#':
                no.address = str(asset) + ':' + str(prop) + ':' + trigger[1:]
                no.trigger = 'change'
                
            #This is an 'alarm activation' trigger...
            if trigger[0] == 'A':
                no.address = 'a:' + str(prop)
                no.trigger = 'alarm'
                
            self.notices.append(no)
            chan = channel()
            chan.ardi = no.address
            self.channels.append(chan)

        #print str(self.notices)


    def ProcessNewValues(self,content,params):
        #print "Processing Content: " + str(content)
        #print str(self.notices)
        for no in self.notices:
            if no.address in content:
                value = content[no.address]
                if no.previous is None:
                    no.previous = value
                else:
                    #print "Trigger: " + str(no.trigger)
                    if no.trigger == 'change':
                        if value != no.previous:
                            self.QueueNotice(no,value)
                            #SendNotice(self.server,no.id,value)
                        no.previous = value
                    if no.trigger == 'alarm':
                        if value != no.previous:
                            self.QueueNotice(no,value)
                            #SendNotice(self.server,no.id,value)
                        no.previous = value

    def QueueNotice(self,notice,value):
        #if notice.delay == 0:
        #    SendNotice(self.server,notice.id,value)
        #else:
            for ni in self.queue:
                if ni.id == notice.id:
                    #Reset timing and leave
                    if ni.type == 'alarm':
                        #This has double-toggled. Ignore and remove.
                        self.queue.remove(ni)
                        pass
                    else:
                        #This has changed again. Don't sent the previous message.
                        ni.timer = time.time();
                        ni.value = value
                    return
            d = Notification()
            d.id = notice.id
            d.delay = notice.delay
            d.timer = time.time()
            d.type = notice.trigger
            d.value = value
            with self.queuelock:
                self.queue.append(d)

    def ProcessQueue(self):
        #print "Checking For Delayed Notifications"
        removal = []

        with self.queuelock:
            for q in self.queue:
                diff = time.time() - q.timer
                if diff > q.delay:
                    removal.append(q)
                    SendNotice(self.server,q.id,q.value)
        with self.queuelock:           
            for q in removal:
                self.queue.remove(q)


    def RunQueueThread(self):
        self.runthread = threading.Thread(target=processqueue)
        self.runthread.daemon = True
        self.runthread.start()
        pass

    def Stop(self):
        super(notifier,self).Stop()
        self.Sub.Disconnect()

note = notifier()
note.signalhandler = CatchSignal
note.Start()
note.RunQueueThread()
note.BlockTillEnd()