#!/usr/bin/python

import signal
import pycurl
import xmltodict
import logging
import platform
from ardi.api import core, subscription
from StringIO import StringIO
from subprocess import call
import subprocess
import os
import traceback
import threading
import time
import argparse

if platform.system() != "Windows":
    from daemon import runner

def CatchSignal(a,b):
    print "Attempting to Terminate"
    signal.signal(signal.SIGINT,signal.SIG_DFL)
    signal.signal(signal.SIGTERM,signal.SIG_DFL)
    note.Stop()

def processqueue():
    while True:
        note.ProcessQueue();
        time.sleep(1)
    

def remove_html_markup(s):
    tag = False
    quote = False
    out = ""

    for c in s:
            if c == '<' and not quote:
                tag = True
            elif c == '>' and not quote:
                tag = False
            elif (c == '"' or c == "'") and tag:
                quote = not quote
            elif not tag:
                out = out + c

    return out

def CallIt(code,params):
    print "    " + code
    cmdline = "send" + code
    env = {}
    for key in params:
        env[key.upper().encode('ascii','ignore')] = str(params[key]).encode('ascii','ignore')

    rt = os.environ.get('SystemRoot','')
    if rt <> "":
        env['SystemRoot'] = rt

    #logging.debug("Sending Via " + cmdline + "...")
    
    #print str(env)
    try:        
        if platform.system() == 'Windows':                
            DETACHED_PROCESS = 0x00000008
            cmdline = os.path.dirname(os.path.realpath(__file__)) + "\\" + cmdline
            #print "Calling " + str(['c:\\python27\\python.exe',cmdline + ".py"])
            subprocess.Popen(['c:\\python27\\python.exe',cmdline + ".py"],env=env, shell=False, stdin=None, stdout=None,stderr=None,close_fds=True,creationflags=DETACHED_PROCESS)
            #call(['c:\\python27\\python.exe',cmdline + ".py"],env=env)
        else:
            cmdline += "d"
            subprocess.call([cmdline],env=env)
    except:
        traceback.print_exc()
        print "Failed To Run Notification Process"

def SendNotice(server,noti,value):
    response = StringIO()

    fullurl = server + "/api/notifier/getnoticeinfo?id=" + str(noti) + "&value=" + str(value)
    #print fullurl

    c = pycurl.Curl()
    c.setopt(c.URL,fullurl)
    c.setopt(c.WRITEFUNCTION,response.write)
    c.setopt(c.TIMEOUT,5)
    c.setopt(c.CONNECTTIMEOUT,5)
    c.perform()
    ret = response.getvalue()
    c.close()

    #print ret

    result = xmltodict.parse(ret)

    #print str(result)

    try:
        methset = result['noticeinfo']['methods']['method']
    except:
        #print "Failure: " + str(methset)
        pass

    methodparams = {}
    #print str(methset)

    if '@code' in methset:
        iset = []
        iset.append(result['noticeinfo']['methods']['method'])
    else:
        iset = result['noticeinfo']['methods']['method']

    #print str(iset)

    for d in iset:
        dta = {}
            
        code = d['@code']
        
        for key in d:
            if (key == '@code'):
                continue
            dta[key[1:]] = str(d[key])

        methodparams[str(code)] = dta

    try:
        methset = result['noticeinfo']['contact']
    except:
        return

    try:
        if '@id' in methset:
            iset = []
            iset.append(result['noticeinfo']['contact'])
        else:
            iset = result['noticeinfo']['contact']
    except:
        print "No Contacts For This Notification"
        return
        
    for d in iset:
        id = d['@id']
        name = d['@name']

        print "Contact " + str(name) + " is contactable via..."

        try:
            if '@code' in d['method']:
                dset = []
                dset.append(d['method'])
            else:
                dset = d['method']
        except:
            #print "Ignoring " + name + " ( " + id + " ) - No Contact Methods"
            continue

        for r in dset:
            dta = {}
            
            code = r['@code']
            #print "    " + code
            for key in r:
                if (key == '@code'):
                    continue
                #print "      " + str(key[1:]) + "=" + str(r[key])
                dta[key[1:]] = str(r[key])

            final = {}
            final.update(dta)
            final.update(methodparams[code])
            final['title'] = result['noticeinfo']['notice']["name"]
            final['detail'] = result['noticeinfo']['notice']["desc"]
            final['footer'] = result['noticeinfo']['notice']["footer"]

            CallIt(code,final)

def CallMe(content,context):
    #print str(content)
    context.ProcessNewValues(content)

class Notification:
    def __init__(self):
        self.style = 'change'
        self.id = 0
        self.timer = 0
        self.value = 0
        self.delay = 0

class notice:
    def __init__(self):
        self.address=''
        self.trigger='#'
        self.delay = 0

class notifier:
    def __init__(self):
        self.notices = []
        #self.initlogger()
        self.methods = {}
        self.queue = []
        self.queuelock = threading.Lock()

    def initlogger(self):

        """Initialises the historian logger"""
        logging.basicConfig(level=logging.INFO)
        ##The logger used to record information about the driver process as a whole
        self.logger = logging.getLogger(__name__)
        
        fname = "/var/log/ardi/notifier-" + self.instancename + ".log"
        if (platform.system() == 'Windows'):
            fname = "c:\\windows\\temp\\ardi-notifier-" + self.instancename + ".log"

        loghandler = logging.FileHandler(fname)
        loghandler.setLevel(logging.DEBUG)

        # create a logging format
        
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        loghandler.setFormatter(formatter)
        
        # add the handlers to the logger

        logging.getLogger("urllib3").setLevel(logging.WARNING)
        logging.getLogger("requests").setLevel(logging.WARNING)
        
        self.logger.addHandler(loghandler)

        self.logger.info('----NOTIFIER RESTART----')
        self.logger.info('Starting Notifier Process')
        
        #redir = iologger(self.logger,logging.ERROR)
        #sys.stderr = redir

    def LoadFromServer(self):
        self.initlogger()
        while True:
            try:
                 servername = self.server
                 response = StringIO()
                
                 fullurl = "http://" + servername + "/api/notifier/getnotifications?trusted=true"
                 #fullurl = "http://" + servername + "/"
                 #print fullurl
                 c = pycurl.Curl()
                 c.setopt(c.URL,fullurl)
                 c.setopt(c.WRITEFUNCTION,response.write)
                 c.setopt(c.TIMEOUT,5)
                 c.setopt(c.CONNECTTIMEOUT,5)
                 c.perform()
                 ret = response.getvalue()
                 c.close()
                 break
            except:
                 time.sleep(10)
                 pass

        #print "Query Complete - '" + ret + "'"
        self.parseconfig(response.getvalue())

    def parseconfig(self,xml):
        """Imports the configuration XML file

        This function loads the list of data points within the XML text provided. The function is usually called from LoadConfig

        \param xml: The configuration data in XML format"""
        result = xmltodict.parse(xml)
                        
        #Core Parameters
        #print str(result)
        
        try:
            if '@name' in result['notices']['notice']:
                pass
        except:
            logging.error('The server currently has no notifications')
            return

        signal.signal(signal.SIGINT,CatchSignal)
        signal.signal(signal.SIGTERM,CatchSignal)

        if '@name' in result['notices']['notice']:
            #print 'Single Item: ' + str(result['notices']['notice'])
            no = notice()
            no.previous = None
            no.id = result['notices']['notice']['@id']

            asset = result['notices']['notice']['@asset']
            prop = result['notices']['notice']['@property']
            trigger = str(result['notices']['notice']['@trigger'])

            #print "Loaded Trigger: " + trigger

            #This is a 'change in value' trigger...
            if trigger[0] == '#':
                no.address = str(asset) + ':' + str(prop) + trigger
                
            #This is an 'alarm activation' trigger...
            if trigger[0] == 'A':
                no.address = 'a:' + str(prop)
                
            #no.trigger = str(result['notices']['notice']['@trigger'])
            no.trigger = 'change'
            self.notices.append(no)
            pass
        else:
            for itm in result['notices']['notice']:
                no = notice()
                no.previous = None
                no.id = itm['@id']

                asset = itm['@asset']
                prop = itm['@property']
                trigger = str(itm['@trigger'])

                #This is a 'change in value' trigger...
                if trigger[0] == '#':
                    no.address = str(asset) + ':' + str(prop) + ':' + trigger[1:]
                    no.trigger = 'change'
                    
                #This is an 'alarm activation' trigger...
                if trigger[0] == 'A':
                    no.address = 'a:' + str(prop)
                    no.trigger = 'alarm'
                    
                #no.trigger = str(result['notices']['notice']['@trigger'])
                #no.trigger = 'change'
                self.notices.append(no)
                pass
                #print 'List: '+ str(itm)

        #print str(self.notices)

    def SetServer(self,ServerName):
        self.server = ServerName

    def ProcessNewValues(self,content):
        #print "Processing Content: " + str(content)
        #print str(self.notices)
        for no in self.notices:
            if no.address in content:
                value = content[no.address]
                if no.previous is None:
                    no.previous = value
                else:
                    #print "Trigger: " + str(no.trigger)
                    if no.trigger == 'change':
                        if value != no.previous:
                            self.QueueNotice(no,value)
                            #SendNotice(self.server,no.id,value)
                        no.previous = value
                    if no.trigger == 'alarm':
                        if value != no.previous:
                            self.QueueNotice(no,value)
                            #SendNotice(self.server,no.id,value)
                        no.previous = value

    def QueueNotice(self,notice,value):
        #if notice.delay == 0:
        #    SendNotice(self.server,notice.id,value)
        #else:
            for ni in self.queue:
                if ni.id == notice.id:
                    #Reset timing and leave
                    if ni.type == 'alarm':
                        #This has double-toggled. Ignore and remove.
                        self.queue.remove(ni)
                        pass
                    else:
                        #This has changed again. Don't sent the previous message.
                        ni.timer = time.time();
                        ni.value = value
                    return
            d = Notification()
            d.id = notice.id
            d.delay = notice.delay
            d.timer = time.time()
            d.type = notice.trigger
            d.value = value
            with self.queuelock:
                self.queue.append(d)
        #pass

    def StartWatching(self):
        Cr = core.ARDICore(self.server)
        Cr.Connect()

        self.Sub = subscription.ARDISubscription(Cr)
        for no in self.notices:
            print '  Listening To ' + no.address
            self.Sub.AddCode(no.address)

        self.Sub.SetCallback(CallMe,self)
        try:
            self.Sub.Connect()
        except:
            traceback.print_exc()

    def ProcessQueue(self):
        #print "Checking For Delayed Notifications"
        removal = []

        with self.queuelock:
            for q in self.queue:
                diff = time.time() - q.timer
                if diff > q.delay:
                    removal.append(q)
                    SendNotice(self.server,q.id,q.value)
        with self.queuelock:           
            for q in removal:
                self.queue.remove(q)


    def RunQueueThread(self):
        self.runthread = threading.Thread(target=processqueue)
        self.runthread.daemon = True
        self.runthread.start()
        pass

    def Stop(self):
        self.Sub.Disconnect()

parser = argparse.ArgumentParser(description='Handle message notification from ARDI')
if platform.system() != "Windows":
    parser.add_argument('service', help='The instruction you would like to give the service')
parser.add_argument('server', help='The ARDI server/site URL to use')

args = parser.parse_args()

note = notifier()
note.SetServer(args.server)
note.instancename = "default"
ps = args.server.rfind('/')
note.instancename = args.server[ps+1:]

#print "Instance: " + note.instancename

if platform.system() == 'Windows':
    #Daemonise this...
    note.LoadFromServer()
    note.RunQueueThread()
    note.StartWatching()
    pass
else:
    #Just run - no need to daemonise.
    #print "Trying to daemonise this thing!"
        
    class Daemon:
        def __init__(self,inst="default"):
            self.stdin_path = '/dev/null'
            self.stdout_path = '/var/log/ardi/notifier-' + inst
            self.stderr_path = '/var/log/ardi/notifier-' + inst
            self.pidfile_path =  '/var/run/ardi-notifier-' + inst + '.pid'
            self.pidfile_timeout = 5
            
        def run(self):
            note.LoadFromServer()
            note.RunQueueThread()
            note.StartWatching()
            
    if args.server != "standalone":
        dm = Daemon(note.instancename)
        dm.note = note
        daemon_runner = runner.DaemonRunner(dm)
        daemon_runner.do_action()
    else:
        note.LoadFromServer()
        note.RunQueueThread()
        note.StartWatching()
    pass